/*
 * cl_image_processor.cpp - CL image processor
 *
 *  Copyright (c) 2015 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author: Wind Yuan <feng.yuan@intel.com>
 */
#include "cl_image_processor.h"
#include "cl_context.h"
#include "cl_device.h"
#include "cl_image_handler.h"
#include "drm_display.h"
#include "cl_demo_handler.h"
#include "xcam_thread.h"

namespace XCam {

class CLHandlerThread
    : public Thread
{
public:
    CLHandlerThread (CLImageProcessor *processor)
        : Thread ("CLHandlerThread")
        , _processor (processor)
    {}
    ~CLHandlerThread () {}

    virtual bool loop ();

private:
    CLImageProcessor *_processor;
};

bool CLHandlerThread::loop ()
{
    XCAM_ASSERT (_processor);
    XCamReturn ret = _processor->process_cl_buffer_queue ();
    if (ret != XCAM_RETURN_NO_ERROR && ret != XCAM_RETURN_BYPASS)
        return false;
    return true;
}

CLImageProcessor::CLImageProcessor (const char* name)
    : ImageProcessor (name ? name : "CLImageProcessor")
    , _seq_num (0)
{
    _context = CLDevice::instance ()->get_context ();
    XCAM_ASSERT (_context.ptr());

    _handler_thread = new CLHandlerThread (this);
    XCAM_ASSERT (_handler_thread.ptr ());

    XCAM_LOG_DEBUG ("CLImageProcessor constructed");
    XCAM_OBJ_PROFILING_INIT;
}

CLImageProcessor::~CLImageProcessor ()
{
    XCAM_LOG_DEBUG ("CLImageProcessor destructed");
}

bool
CLImageProcessor::add_handler (SmartPtr<CLImageHandler> &handler)
{
    XCAM_ASSERT (handler.ptr ());
    _handlers.push_back (handler);
    return true;
}

SmartPtr<CLContext>
CLImageProcessor::get_cl_context ()
{
    return _context;
}

bool
CLImageProcessor::can_process_result (SmartPtr<X3aResult> &result)
{
    XCAM_UNUSED (result);
    return false;
}

XCamReturn
CLImageProcessor::apply_3a_results (X3aResultList &results)
{
    XCAM_UNUSED (results);
    return XCAM_RETURN_NO_ERROR;
}

XCamReturn
CLImageProcessor::apply_3a_result (SmartPtr<X3aResult> &result)
{
    XCAM_UNUSED (result);
    return XCAM_RETURN_NO_ERROR;
}

XCamReturn
CLImageProcessor::process_buffer (SmartPtr<VideoBuffer> &input, SmartPtr<VideoBuffer> &output)
{
    SmartPtr<DrmBoBuffer> drm_bo_in;
    XCamReturn ret = XCAM_RETURN_NO_ERROR;
    SmartPtr<DrmDisplay> display = DrmDisplay::instance ();

    drm_bo_in = display->convert_to_drm_bo_buf (display, input);
    XCAM_FAIL_RETURN (
        WARNING,
        drm_bo_in.ptr (),
        XCAM_RETURN_ERROR_MEM,
        "CL image processor can't handle this buffer, maybe type error");

    while (!_done_buffer_queue.is_empty ()) {
        SmartPtr<DrmBoBuffer> done_buf = _done_buffer_queue.pop (50000); //50ms
        if (!done_buf.ptr ())
            break;
        //notify buffer done
        notify_process_buffer_done (done_buf);
    }
    output = NULL; // consider call back?

    STREAM_LOCK;

    if (_handlers.empty()) {
        ret = create_handlers ();
    }

    XCAM_FAIL_RETURN (
        WARNING,
        !_handlers.empty () && ret == XCAM_RETURN_NO_ERROR,
        XCAM_RETURN_ERROR_CL,
        "CL image processor create handlers failed");

    SmartPtr<PriorityBuffer> p_buf = new PriorityBuffer;
    p_buf->set_seq_num (_seq_num++);
    p_buf->data = drm_bo_in;
    p_buf->handler = *(_handlers.begin ());

    XCAM_FAIL_RETURN (
        WARNING,
        _process_buffer_queue.push_priority_buf (p_buf),
        XCAM_RETURN_ERROR_UNKNOWN,
        "CLImageProcessor push priority buffer failed");

    return XCAM_RETURN_BYPASS;
}

XCamReturn
CLImageProcessor::process_cl_buffer_queue ()
{
    XCamReturn ret = XCAM_RETURN_NO_ERROR;
    SmartPtr<PriorityBuffer> p_buf = _process_buffer_queue.pop (-1);
    if (!p_buf.ptr ()) {
        XCAM_LOG_DEBUG ("cl buffer queue stopped");
        return XCAM_RETURN_ERROR_MEM;
    }

    SmartPtr<DrmBoBuffer> data = p_buf->data;
    SmartPtr<CLImageHandler> handler = p_buf->handler;
    SmartPtr <DrmBoBuffer> out_data;

    XCAM_ASSERT (data.ptr () && handler.ptr ());

    XCAM_LOG_DEBUG ("buf:%d, rank:%d\n", p_buf->seq_num, p_buf->rank);

    {
        STREAM_LOCK;
        ret = handler->execute (data, out_data);
        XCAM_FAIL_RETURN (
            WARNING,
            ret == XCAM_RETURN_NO_ERROR,
            ret,
            "CLImageProcessor execute image handler failed");
        XCAM_ASSERT (out_data.ptr ());

        // for loop in handler, find next handler
        ImageHandlerList::iterator i_handler = _handlers.begin ();
        for (; i_handler != _handlers.end ();  ++i_handler)
        {
            if (handler.ptr () == (*i_handler).ptr ()) {
                ++i_handler;
                if (i_handler != _handlers.end ())
                    p_buf->handler = *i_handler;
                else
                    p_buf->handler = NULL;
                break;
            }
        }
    }

    // buffer processed by all handlers, done
    if (!p_buf->handler.ptr ()) {
        if (out_data.ptr ())
            out_data->clear_attached_buffers ();

        XCAM_OBJ_PROFILING_START;
        CLDevice::instance()->get_context ()->finish ();
        XCAM_OBJ_PROFILING_END (get_name (), 30);

        // buffer done, push back
        _done_buffer_queue.push (out_data);
        return XCAM_RETURN_NO_ERROR;
    }

    p_buf->data = out_data;
    p_buf->down_rank ();

    XCAM_FAIL_RETURN (
        WARNING,
        _process_buffer_queue.push_priority_buf (p_buf),
        XCAM_RETURN_ERROR_UNKNOWN,
        "CLImageProcessor push priority buffer failed");

    return ret;
}

XCamReturn
CLImageProcessor::emit_start ()
{
    _done_buffer_queue.resume_pop ();
    _process_buffer_queue.resume_pop ();

    if (!_handler_thread->start ())
        return XCAM_RETURN_ERROR_THREAD;

    return XCAM_RETURN_NO_ERROR;
}

void
CLImageProcessor::emit_stop ()
{
    _process_buffer_queue.pause_pop();
    _done_buffer_queue.pause_pop ();


    for (ImageHandlerList::iterator i_handler = _handlers.begin ();
            i_handler != _handlers.end ();  ++i_handler) {
        (*i_handler)->emit_stop ();
    }

    _handler_thread->stop ();
    _process_buffer_queue.clear ();
    _done_buffer_queue.clear ();
}

XCamReturn
CLImageProcessor::create_handlers ()
{
    SmartPtr<CLImageHandler> demo_handler;
    demo_handler = create_cl_demo_image_handler (_context);
    XCAM_FAIL_RETURN (
        WARNING,
        demo_handler.ptr (),
        XCAM_RETURN_ERROR_CL,
        "CLImageProcessor create demo handler failed");
    add_handler (demo_handler);

    return XCAM_RETURN_NO_ERROR;
}

};
